using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Security.Cryptography;
using System.Threading;

namespace DupFinder
{
    /// <summary>
    /// http://www.kerrywong.com
    /// </summary>
    public class DupFinderCore2
    {
        private Hashtable _hash = new Hashtable();
        private AutoResetEvent[] _autoEvents;
        private object SyncObject = new Object();

        private const int _numOfThreads = 10;

        public Hashtable Hash
        {
            get
            {
                lock (SyncObject)
                {
                    return _hash;
                }
            }
            set
            {
                lock (SyncObject)
                {
                    _hash = value;
                }
            }
        }

        public void FindDuplicates(List<string> files)
        {
            for (int n = 0; n < files.Count; n += _numOfThreads)
            {
                _autoEvents = new AutoResetEvent[_numOfThreads];

                for (int i = 0; i < _numOfThreads; i++)
                {
                    if (n + i < files.Count)
                    {
                        _autoEvents[i] = new AutoResetEvent(false);
                        ThreadPool.QueueUserWorkItem(CalculateFileHash, new FileThreadInfo(i, files[n + i]));
                    }
                    else
                    {
                        _autoEvents[i] = new AutoResetEvent(true);
                    }
                }

                WaitHandle.WaitAll(_autoEvents);
            }
        }

        private void CalculateFileHash(object stateInfo)
        {
            FileThreadInfo threadInfo = stateInfo as FileThreadInfo;
            string h = ComputeMD5(threadInfo.FileName);

            Thread.CurrentThread.Name = threadInfo.FileName;

            if (h != string.Empty)
            {
                FileDataDefinition def = new FileDataDefinition();
                def.FileName = Path.GetFileName(threadInfo.FileName);
                def.FilePath = Path.GetDirectoryName(threadInfo.FileName);
                def.LastModifiedDate = Directory.GetLastWriteTime(threadInfo.FileName);

                FileInfo fileInfo = new FileInfo(threadInfo.FileName);
                def.FileSize = fileInfo.Length;

                if (Hash.Contains(string.Concat(def.FileName, ":", h)))
                {
                    List<FileDataDefinition> l = Hash[string.Concat(def.FileName, ":", h)] as List<FileDataDefinition>;
                    l.Add(def);

                    Hash[string.Concat(def.FileName, ":", h)] = l;
                }
                else
                {
                    List<FileDataDefinition> l = new List<FileDataDefinition>();
                    l.Add(def);
                    Hash.Add(string.Concat(def.FileName, ":", h), l);
                }
            }

            _autoEvents[threadInfo.ID].Set();
        }

        public string ComputeMD5(string fileName)
        {
            MD5 md5 = new MD5CryptoServiceProvider();
            byte[] hash = null;

            FileStream fs = null;

            try
            {
                fs = new FileStream(fileName, FileMode.Open);
                hash = md5.ComputeHash(fs);
            }
            catch
            {
            }
            finally
            {
                if (fs != null)
                {
                    fs.Close();
                }
            }

            if (hash != null)
                return Convert.ToBase64String(hash);
            else
                return string.Empty;

        }

    }
}
