using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Threading;
using System.Security.Cryptography;

namespace DupFinder
{
    /// <summary>
    /// http://www.kerrywong.com
    /// </summary>
    public class FileBytes
    {
        public FileDataDefinition fileDefinition;
        public byte[] bytes;

        public FileBytes(FileDataDefinition fileDefinition, byte[] bytes)
        {
            this.fileDefinition = fileDefinition;
            this.bytes = bytes;
        }
    }

    //Andrew D. Birrell 
    //http://research.microsoft.com/~birrell/papers/ThreadsCSharp.pdf
    public class ProducerConsumerQueue : IDisposable
    {
        private Hashtable _hash = new Hashtable();

        public Hashtable Hash
        {
            get
            {
                return _hash;
            }
            set
            {
                _hash = value;
            }
        }

        EventWaitHandle wh = new AutoResetEvent(false);
        Thread worker;
        object locker = new object();
        Queue<FileBytes> tasks = new Queue<FileBytes>();
        public ProducerConsumerQueue()
        {
            worker = new Thread(Work);
            worker.Start();
        }
        public void EnqueueTask(FileBytes task)
        {
            lock (locker) tasks.Enqueue(task);
            wh.Set();
        }
        public void Dispose()
        {
            EnqueueTask(null); // Signal the consumer to exit.
            worker.Join(); // Wait for the consumer's thread to finish.
            wh.Close(); // Release any OS resources.
        }

        void Work()
        {
            FileDataDefinition def = null;
            MD5 md5 = new MD5CryptoServiceProvider();

            while (true)
            {
                FileBytes task = null;
                lock (locker)
                    if (tasks.Count > 0)
                    {
                        task = tasks.Dequeue();
                        
                        if (task == null) return;
                        def = task.fileDefinition;
                    }
                if (task != null)
                {

                    string h = Convert.ToBase64String(md5.ComputeHash(task.bytes));

                    if (Hash.Contains(string.Concat(def.FileName, ":", h)))
                    {
                        List<FileDataDefinition> l = Hash[string.Concat(def.FileName, ":", h)] as List<FileDataDefinition>;
                        l.Add(def);

                        Hash[string.Concat(def.FileName, ":", h)] = l;
                    }
                    else
                    {
                        List<FileDataDefinition> l = new List<FileDataDefinition>();
                        l.Add(def);
                        Hash.Add(string.Concat(def.FileName, ":", h), l);
                    }
                }
                else
                    wh.WaitOne(); // No more tasks - wait for a signal
            }
        }

        public void OutputDuplicates()
        {
            foreach (DictionaryEntry entry in Hash)
            {
                List<FileDataDefinition> l = entry.Value as List<FileDataDefinition>;

                if (l.Count > 1)
                {
                    string s = string.Empty;

                    foreach (FileDataDefinition def in l)
                    {
                        s = string.Concat(s, def.FilePath, "\t");
                    }

                    Console.WriteLine("{0} \t {1} \t {2} \t {3}", l.Count, l[0].FileName, l[0].FileSize, s);
                }
            }
        }
    }

    public class DupFinderCore3
    {
        public ProducerConsumerQueue _q = null;

        public void FindDuplicates(string rootDir)
        {
            using (_q = new ProducerConsumerQueue())
            {
                GetFiles(rootDir);
            }
            _q.OutputDuplicates();
        }

        public void GetFiles(string rootDir)
        {
            string[] files = null;

            try
            {
                files = Directory.GetFiles(rootDir);

                foreach (string fn in files)
                {
                    FileDataDefinition def = new FileDataDefinition();
                    def.FileName = Path.GetFileName(fn);
                    def.FilePath = Path.GetDirectoryName(fn);
                    def.LastModifiedDate = Directory.GetLastWriteTime(fn);
                    
                    FileInfo fileInfo = new FileInfo(fn);
                    def.FileSize = fileInfo.Length;

                    FileStream f = new FileStream(fn, FileMode.Open);
                    byte[] bytes = new byte[f.Length];                   
                    f.Read(bytes, 0, (int) f.Length - 1);
                    f.Close();

                    _q.EnqueueTask(new FileBytes(def, bytes));
                }
            }
            catch (Exception e)
            {
            }
            string[] dirs = null;

            try
            {
                dirs = Directory.GetDirectories(rootDir);

                foreach (string s in dirs)
                {
                    GetFiles(s);
                }
            }
            catch
            {
            }
        }
    }
}